@@ -23,17 +23,22 @@
 #ifndef GRD_TYPES_H
 #define GRD_TYPES_H
 
-typedef struct _GrdContext GrdContext;
 typedef struct _GrdClipboard GrdClipboard;
 typedef struct _GrdClipboardRdp GrdClipboardRdp;
 typedef struct _GrdClipboardVnc GrdClipboardVnc;
+typedef struct _GrdContext GrdContext;
+typedef struct _GrdCredentials GrdCredentials;
 typedef struct _GrdEglThread GrdEglThread;
 typedef struct _GrdHwAccelNvidia GrdHwAccelNvidia;
-typedef struct _GrdRdpDamageDetector GrdRdpDamageDetector;
-typedef struct _GrdRdpDisplayControl GrdRdpDisplayControl;
-typedef struct _GrdRdpEventQueue GrdRdpEventQueue;
+typedef struct _GrdRdpAudioOutputStream GrdRdpAudioOutputStream;
+typedef struct _GrdRdpAudioPlayback GrdRdpAudioPlayback;
+typedef struct _GrdRdpAudioVolumeData GrdRdpAudioVolumeData;
 typedef struct _GrdRdpBuffer GrdRdpBuffer;
 typedef struct _GrdRdpBufferPool GrdRdpBufferPool;
+typedef struct _GrdRdpDamageDetector GrdRdpDamageDetector;
+typedef struct _GrdRdpDisplayControl GrdRdpDisplayControl;
+typedef struct _GrdRdpDvc GrdRdpDvc;
+typedef struct _GrdRdpEventQueue GrdRdpEventQueue;
 typedef struct _GrdRdpGfxFrameController GrdRdpGfxFrameController;
 typedef struct _GrdRdpGfxFrameLog GrdRdpGfxFrameLog;
 typedef struct _GrdRdpGfxSurface GrdRdpGfxSurface;
@@ -42,6 +47,7 @@ typedef struct _GrdRdpNetworkAutodetection GrdRdpNetworkAutodetection;
 typedef struct _GrdRdpSAMFile GrdRdpSAMFile;
 typedef struct _GrdRdpServer GrdRdpServer;
 typedef struct _GrdRdpSurface GrdRdpSurface;
+typedef struct _GrdRdpTelemetry GrdRdpTelemetry;
 typedef struct _GrdSession GrdSession;
 typedef struct _GrdSessionRdp GrdSessionRdp;
 typedef struct _GrdSessionVnc GrdSessionVnc;
